;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                                          ;;
;;         "BootProg" Loader v 1.5 by Alexey Frunze (c) 2000-2015           ;;
;;                           2-clause BSD license.                          ;;
;;                                                                          ;;
;;                                                                          ;;
;;                              How to Compile:                             ;;
;;                              ~~~~~~~~~~~~~~~                             ;;
;; nasm bootex.asm -f bin -o bootex.bin                                     ;;
;;                                                                          ;;
;;                                                                          ;;
;;                                 Features:                                ;;
;;                                 ~~~~~~~~~                                ;;
;; - exFAT supported using BIOS int 13h function 42h.                       ;;
;;                                                                          ;;
;; - Loads a 16-bit executable file in the MS-DOS .COM or .EXE format       ;;
;;   from the root directory of a disk and transfers control to it          ;;
;;   (the "ProgramName" variable holds the name of the file to be loaded)   ;;
;;   Its maximum size can be up to 637KB without Extended BIOS Data area.   ;;
;;                                                                          ;;
;; - Prints an error if the file isn't found or couldn't be read            ;;
;;   ("File not found" or "Read error")                                     ;;
;;   and waits for a key to be pressed, then executes the Int 19h           ;;
;;   instruction and lets the BIOS continue bootstrap.                      ;;
;;                                                                          ;;
;;                                                                          ;;
;;                                Known Bugs:                               ;;
;;                                ~~~~~~~~~~~                               ;;
;; - All bugs are fixed as far as I know. The boot sector has been tested   ;;
;;   on a 128MB qemu image.                                                 ;;
;;                                                                          ;;
;;                                                                          ;;
;;                               Memory Layout:                             ;;
;;                               ~~~~~~~~~~~~~~                             ;;
;; The diagram below shows the typical memory layout. The actual location   ;;
;; of the boot sector and its stack may be lower than A0000H if the BIOS    ;;
;; reserves memory for its Extended BIOS Data Area just below A0000H and    ;;
;; reports less than 640 KB of RAM via its Int 12H function.                ;;
;;                                                                          ;;
;;                                            physical address              ;;
;;                 +------------------------+ 00000H                        ;;
;;                 | Interrupt Vector Table |                               ;;
;;                 +------------------------+ 00400H                        ;;
;;                 |     BIOS Data Area     |                               ;;
;;                 +------------------------+ 00500H                        ;;
;;                 | PrtScr Status / Unused |                               ;;
;;                 +------------------------+ 00600H                        ;;
;;                 |      Loaded Image      |                               ;;
;;                 +------------------------+ nnnnnH                        ;;
;;                 |    Available Memory    |                               ;;
;;                 +------------------------+ A0000H - 1KB                  ;;
;;                 |       Boot Sector      |                               ;;
;;                 +------------------------+ A0000H - 0.5KB                ;;
;;                 |    0.5KB Boot Stack    |                               ;;
;;                 +------------------------+ A0000H                        ;;
;;                 |        Video RAM       |                               ;;
;;                                                                          ;;
;;                                                                          ;;
;;                   Boot Image Startup (register values):                  ;;
;;                   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~                  ;;
;;  ax = 0ffffh (both FCB in the PSP don't have a valid drive identifier),  ;;
;;  bx = 0, dl = BIOS boot drive number (e.g. 0, 80H)                       ;;
;;  cs:ip = program entry point                                             ;;
;;  ss:sp = program stack (don't confuse with boot sector's stack)          ;;
;;  COM program defaults: cs = ds = es = ss = 50h, sp = 0, ip = 100h        ;;
;;  EXE program defaults: ds = es = EXE data - 10h (fake MS-DOS psp),       ;;
;;  cs:ip and ss:sp depends on EXE header                                   ;;
;;  Magic numbers:                                                          ;;
;;    si = 16381 (prime number 2**14-3)                                     ;;
;;    di = 32749 (prime number 2**15-19)                                    ;;
;;    bp = 65521 (prime number 2**16-15)                                    ;;
;;  The magic numbers let the program know whether it has been loaded by    ;;
;;  this boot sector or by MS-DOS, which may be handy for universal, bare-  ;;
;;  metal and MS-DOS programs.                                              ;;
;;  The command line contains no arguments.                                 ;;
;;                                                                          ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

%define bx(label)       bx+label-boot
%define si(label)       si+label-boot
NullEntryCheck          equ     1               ; +3/5 bytes
ReadRetry               equ     1               ; +7 bytes
SectorOf512Bytes        equ     0               ; -12 bytes
CheckAttrib             equ     0               ; +16/18 bytes
WaitForKey              equ     0               ; +5 bytes
TfatSupport             equ     1               ; +10 bytes
CheckLBAsupport         equ     0               ; +11/21 bytes

[BITS 16]
[CPU 386]

ImageLoadSeg            equ     60h
StackSize               equ     512

[SECTION .text]
[ORG 0]

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Boot sector starts here ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

boot:
DriveNumber:
        jmp     short   start                   ; Windows checks for this jump
        nop
bsOemName               times   8  db " "       ; 0x03 "EXFAT   "
                        times   53 db 0         ; 0x0B

;;;;;;;;;;;;;;;;;;;;;
;; BPB starts here ;;
;;;;;;;;;;;;;;;;;;;;;

bpbSectorStart          DQ      0               ; 0x40 partition first sector
bpbSectorCount          DQ      0               ; 0x48 partition sectors count
bpbFatSectorStart       DD      0               ; 0x50 FAT first sector
bpbFatSectorCount       DD      0               ; 0x54 FAT sectors count
bpbClusterSectorStart   DD      0               ; 0x58 first cluster sector
bpbClusterCount         DD      0               ; 0x5C total clusters count
bpbRootDirCluster       DD      0               ; 0x60 first cluster of the root dir
bpbVolumeSerial         DD      0               ; 0x64 volume serial number
bpbFSVersionMinor       DB      0               ; 0x68
bpbFSVersionMajor       DB      0               ; 0x69
bpbVolumeStateFlags     DW      0               ; 0x6A bit0 = fat used, bit1 = dirty, bit2 = media error 
bpbSectorSizeBits       DB      0               ; 0x6C sector size as (1 << n)
bpbSectorPerClusterBits DB      0               ; 0x6D sector per cluster as (1 << n)
bpbNumberOfFATs         DB      0               ; 0x6E always 1 or 2 (Tfat case)
bpbDriveNumber          DB      0               ; 0x6F always 0x80
bpbAllocatedPercent     DB      0               ; 0x70 percentage of allocated space

;;;;;;;;;;;;;;;;;;;
;; BPB ends here ;;
;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Boot sector code starts here ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

start:
        cld

;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; How much RAM is there? ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        int     12h             ; get conventional memory size (in KBs)
        mov     cx, 106h
        dec     ax              ; reserve 1K bytes for the code and the stack
        shl     ax, cl          ; and convert it to 16-byte paragraphs

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Reserve memory for the boot sector and its stack ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        mov     es, ax                  ; cs:0 = ds:0 = ss:0 -> top - 512 - StackSize
        mov     ss, ax
        mov     sp, 512+StackSize       ; bytes 0-511 are reserved for the boot code

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Copy ourselves to top of memory ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        mov     si, 7C00h
        xor     di, di
        mov     ds, di
        push    es
        mov     [si(DriveNumber)], dx   ; store BIOS boot drive number
        rep     movsw                   ; move 512 bytes (+ 12)

;;;;;;;;;;;;;;;;;;;;;;
;; Jump to the copy ;;
;;;;;;;;;;;;;;;;;;;;;;

        push    word main
        retf

%if CheckLBAsupport != 0
%macro BootFileName 0
main:
        mov     ah, 41h                 ; clobbers AX, BX, CX, DH
        mov     bx, 55AAh
        int     13h
        jc      ReadError
;       xor     bx, 0AA55h
;       jnz     ReadError
;       shr     cx, 1                   ; function 42h support ?
;       jnc     ReadError
        xor     cx, cx
%else
main:
%endif
        push    cs
        pop     ds

        xor     ebx, ebx

        mov     esi, [bx(bpbRootDirCluster)] ; esi=cluster # of root dir

        push    byte ImageLoadSeg
        pop     es                      ; cx = 0

RootDirReadContinue:
        call    ReadCluster             ; read one sector of root dir

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Look for the COM/EXE file to load and run ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

                                        ; es:di -> root entries array

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Looks for the file/dir ProgramName    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Input:  ES:DI -> root directory array ;;
;; Output: ESI = cluster number          ;;
;;         dword [bx+FileSize] file size ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

CurNameSize     equ  03h                ; 1 byte
StartCluster    equ  14h                ; 4 bytes
FileSize        equ  18h                ; 8 bytes

FindNameCycle:
        pusha

%if NullEntryCheck != 0
        xor     ax, ax
        or      al, [es:di]
        je      FindNameFailed
 %define curtag al 
 %define chktag al 
%else
 %define curtag byte [es:di] 
 %define chktag ax
%endif
        cmp     curtag, 0c0h            ; EXFAT_ENTRY_FILE_INFO ?
        jne     NotFileInfo

        mov     bl, 32
CopyInfo:
        mov     [bx], al
        dec     bx
        mov     al, [es:di+bx]
        jnz     CopyInfo                ; keep BIOS boot drive number

NotFileInfo:
%if CheckAttrib != 0
Attributes      equ  0Bh                ; 1 byte
        cmp     curtag, 85h             ; EXFAT_ENTRY_FILE ?
        jne     NotEntryFile
        mov     al, [es:di+Attributes]
        mov     [ProgramName+NameLength+1], al
NotEntryFile:
%endif
        mov     chktag, 0c1h            ; EXFAT_ENTRY_FILE_NAME ?
        mov     cx, NameLength+1
        mov     si, ProgramName         ; ds:si -> program name
CheckName:
        scasw                           ; compare UTF-16
        lodsb                           ; with ASCII
        loope   CheckName
VolumeLabel     equ  8
SubDirectory    equ  10h
%if CheckAttrib != 0
        jnz     SkipFindName
        test    byte [si], VolumeLabel+SubDirectory
SkipFindName:
%endif
        je      FindNameFound           ; cx = 0
        popa                            ; restore ax, cx, si, di

        add     di, byte 32
        cmp     di, bp
        jne     FindNameCycle           ; next root entry
        loop    RootDirReadContinue     ; continue to the next root dir sector
        cmp     esi, byte -10           ; carry=0 if last cluster, and carry=1 otherwise
        jc      RootDirReadContinue     ; continue to the next root dir cluster
FindNameFailed:                         ; end of root directory (dir end reached)
        mov     dl, [bx(DriveNumber)]   ; restore BIOS boot drive number
        call    Error
        db      "File not found."
FindNameFound:
        mov     esi, [bx+StartCluster]

;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Load the entire file ;;
;; Input: ESI = cluster ;;
;;        CX  = 0       ;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;

        push    es
%if SectorOf512Bytes == 0
        xor     bp, bp
FileReadContinue:
        shr     bp, 4                   ; bytes to paragraphs
        mov     di, es
        add     di, bp                  ; adjust segment for next sector
        mov     es, di                  ; es:0 updated
%else
FileReadContinue:
%endif
        call    ReadCluster             ; read one more sector of the boot file
        dec     cx
        sub     [bx+FileSize], ebp      ; max FileSize is < 640KB : check low 32 bits only
%if SectorOf512Bytes != 0
        mov     bp, es
        lea     bp, [bp+32]
        mov     es, bp                  ; es:0 updated
%endif
        ja      FileReadContinue
        mov     dx, [bx(DriveNumber)]   ; restore BIOS boot drive number
        xchg    ax, di
        pop     bp

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Type detection, .COM or .EXE? ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        mov     ds, bp                  ; bp=ds=seg the file is loaded to

        add     bp, [bx+08h]            ; bp = image base
        mov     di, [bx+18h]            ; di = reloc table pointer

        cmp     word [bx], 5A4Dh        ; "MZ" signature?
        je      RelocateEXE             ; yes, it's an EXE program

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Setup and run a .COM program ;;
;; Set CS=DS=ES=SS SP=0 IP=100h ;;
;; AX=0ffffh BX=0 DX=drive and  ;;
;; cmdline=void                 ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        mov     di, 100h                ; ip
        mov     bp, ImageLoadSeg-10h    ; "org 100h" stuff :)
        mov     ss, bp
        xor     sp, sp
        push    bp                      ; cs, ds and es
        jmp     short Run

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Relocate, setup and run a .EXE program     ;;
;; Set CS:IP, SS:SP, DS, ES and AX according  ;;
;; to wiki.osdev.org/MZ#Initial_Program_State ;;
;; AX=0ffffh BX=0 DX=drive cmdline=void       ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

ReloCycle:
        add     [di+2], bp              ; item seg (abs)
        les     si, [di]                ; si = item ofs, es = item seg
        add     [es:si], bp             ; fixup
        scasw                           ; di += 2
        scasw                           ; point to next entry

RelocateEXE:
        dec     word [bx+06h]           ; reloc items, 32768 max (128KB table)
        jns     ReloCycle

        les     si, [bx+0Eh]
        add     si, bp
        mov     ss, si                  ; ss for EXE
        mov     sp, es                  ; sp for EXE

        lea     si, [bp-10h]            ; ds and es both point to the segment
        push    si                      ; containing the PSP structure

        add     bp, [bx+16h]            ; cs for EXE
        mov     di, [bx+14h]            ; ip for EXE
Run:
        pop     ds
        push    bp
        push    di
        push    ds
        pop     es
        mov     [80h], ax               ; clear cmdline
        dec     ax                      ; both FCB in the PSP don't have a valid drive identifier

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Set the magic numbers so the program knows that it   ;;
;; has been loaded by this bootsector and not by MS-DOS ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
        mov     si, 16381 ; prime number 2**14-3
        mov     di, 32749 ; prime number 2**15-19
        mov     bp, 65521 ; prime number 2**16-15

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; All done, transfer control to the program now ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
        retf
%if CheckLBAsupport != 0
%endm
%endif

;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Error Messaging Code ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;

Error:
        pop     si

PutStr:
        mov     ah, 0Eh
        mov     bl, 7
        lodsb
        int     10h
        cmp     al, "."
        jne     PutStr
%if WaitForKey != 0
        cbw
        int     16h                     ; wait for a key...
        int     19h                     ; bootstrap
%endif
Stop:
        hlt
        jmp     short Stop

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Reads a exFAT cluster         ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Input:  EDX:EAX = LBA         ;;
;;         EBX     = 0           ;;
;;         CX      = sector cnt  ;;
;;         ESI     = cluster no  ;;
;;         ES:0   -> buffer adrs ;;
;; Output: EDX:EAX = next LBA    ;;
;;         CX      = sector cnt  ;;
;;         ESI     = cluster no  ;;
;;         EBP     = bytes/sector;;
;; Keep:   EDI     = 0           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

ReadCluster:
        add     eax, byte 1

        inc     cx                              ; jcxnz
        loop    ReadSectorC

        mul     ebx                             ; edx:eax = 0
%if SectorOf512Bytes != 0
        mov     al, 128
%define SectorPerClusterBits    [bx(bpbSectorPerClusterBits)]
%else
        mov     ah, 40h
        mov     cx, [bx(bpbSectorSizeBits)]
%define SectorPerClusterBits    ch
        rol     ax, cl                          ; eax=# of exFAT entries per sector
%endif
        lea     edi, [esi-2]                    ; edi=cluster #-2
        xchg    eax, esi
        div     esi                             ; eax=FAT sector #, edx=entry # in sector

        imul    si, dx, byte 4                  ; si=entry # offset in sector

        cdq
        add     eax, [bx(bpbFatSectorStart)]    ; sector # relative to exFAT
%if TfatSupport
        test    byte [bx(bpbVolumeStateFlags)], 1
        jz      UseFat0
        add     eax, [bx(bpbFatSectorCount)]
UseFat0:
%endif
        call    ReadSectorFAT                   ; read 1 exFAT sector, keep edx=0, set C

        mov     esi, [es:si]                    ; esi=next cluster #

        mov     dl, SectorPerClusterBits
        xor     ecx, ecx
        bts     ecx, edx                        ; 10000h max (32MB cluster)
        xchg    eax, edi                        ; get cluster #-2
        mul     ecx

        add     eax, [bx(bpbClusterSectorStart)]
ReadSectorC:
        mov     di, bx
ReadSectorFAT:
        adc     edx, ebx

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Reads a sector using BIOS Int 13h ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Input:  EDX:EAX = LBA             ;;
;;         EBX     = 0               ;;
;;         ES:0   -> buffer address  ;;
;; Output: EBP     = bytes/sector    ;;
;; Keep:   ESI     = cluster         ;;
;;         EDI     = FAT sector or 0 ;;
;;         ECX     = sector count    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

%if SectorOf512Bytes != 0
        lea     ebp, [bx+512]
%else
        lea     ebp, [bx+1]
%endif

        pushad

        add     eax, [bx(bpbSectorStart)]
        adc     edx, [bx(bpbSectorStart)+4]

        push    edx
        push    eax
        push    es
        push    bx
%if SectorOf512Bytes != 0
        push    byte 1                  ; sector count word = 1
%else
        push    bp                      ; sector count word = 1
%endif
        push    byte 16                 ; packet size byte = 16, reserved byte = 0
ReadSectorRetry:        
        mov     si, sp
        mov     ah, 42h                 ; ah = 42h = extended read function no.
        mov     dl, [bx(DriveNumber)]   ; restore BIOS boot drive number
        int     13h                     ; extended read sectors (DL, DS:SI)

        jnc     ReadSuccess

%if ReadRetry != 0
        xor     ax, ax
        int     13h                     ; reset drive (DL)
        dec     bp
 %if SectorOf512Bytes != 0
        jne     ReadSectorRetry         ; up to 511 tries
 %else
        jpe     ReadSectorRetry         ; up to 3 tries
 %endif
%endif

ReadError:
        call    Error
        db      "Read error."

ReadSuccess:
%if SectorOf512Bytes == 0
        mov     cl, [bx(bpbSectorSizeBits)]
        shl     word [si+16+8], cl      ; (e)bp si+16: EDI ESI EBP ESP EBX EDX ECX EAX
%endif
        popa                            ; sp += 16
        popad                           ; real registers
        ret

%if CheckLBAsupport != 0
        BootFileName
%endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Fill free space with zeroes ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

                times (512-13-($-$$)) db 0

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Name of the file to load and run ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

NameLength      equ   11
ProgramName     times NameLength db 0   ; name and extension

;;;;;;;;;;;;;;;;;;;;;;;;;;
;; End of the sector ID ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;

                dw      0AA55h          ; BIOS checks for this ID
